<!---+
	Main component for routing, provides facilities for matching and generating
	routes. 
	
	Author: Elliott H. Sprehn
	Rev:    1.5
--->
<cfcomponent output="false" extends="accelerate.Object">

	<!--- Include mixin that provides object cache functionality --->
	<cfset mixin("objectcache")>

	<!---+
		Constructor. Initializes the mapper.
				
		prefix:
			A prefix that should be added to all routes when generating them.
			This is useful when your application is in a nested directory.
			
			Example: "/test"
			
		maxCachedRoute:
			The maximum number of matched and generated routes to cache. This
			should be set with available memory and possible route combinations
			in mind. Should be an integer.
			
		TODO: This method is missing documentation for lots of parameters!
	--->
	<cffunction name="init" access="public" output="false" returntype="any">
		<cfargument name="prefix" type="string" required="false" default="">
		<cfargument name="appendSlash" type="boolean" required="false" default="false">
		<cfargument name="scriptName" type="string" required="false" default="cgi.script_name">
		<cfargument name="absoluteUrl" type="boolean" required="false" default="false">
		<cfargument name="protocol" type="string" required="false" default="auto">
		<cfargument name="serverName" type="string" required="false" default="cgi.server_name">
		<cfargument name="maxCachedRoutes" type="numeric" required="false" default="250">
		<cfargument name="routingDefaults" type="struct" required="false" default="#structNew()#">
		
		<cfset super.init()>		
		
		<cfset variables.namedRoutes = structNew()>
		<cfset variables.routes = arrayNew(1)>
				
		<!--- Generation defaults, see generate() --->
		<cfset setAppendSlash( arguments.appendSlash )>
		<cfset setScriptName( arguments.scriptName )>
		<cfset setAbsoluteUrl( arguments.absoluteUrl )>
		<cfset setProtocol( arguments.protocol )>
		<cfset setServerName( arguments.serverName )>
		<cfset setMaxCachedRoutes( arguments.maxCachedRoutes )>
		<cfset setRoutingDefaults( arguments.routingDefaults )>
		
		<cfset setPrefix( arguments.prefix )>
		
		<cfset setObjectCacheLockName("accelerate.routing.Mapper(#createUUID()#)")>
				
		<cfreturn this>
	</cffunction>
	
	<cffunction name="setRoutingDefaults" access="public" returntype="void" output="false">
		<cfargument name="value" type="struct" required="true">

		<cfset variables.routingDefaults = arguments.value>
	</cffunction>
	
	<cffunction name="getRoutingDefaults" access="public" returntype="any" output="false">

		<cfreturn variables.routingDefaults>
	</cffunction>
	
	<cffunction name="setMaxCachedRoutes" access="public" returntype="void" output="false">
		<cfargument name="value" type="numeric" required="true">

		<cfset variables.maxCachedRoutes = arguments.value>
		<cfset setMaxCachedObjects( arguments.value )>		
	</cffunction>
	
	<cffunction name="getMaxCachedRoutes" access="public" returntype="any" output="false">

		<cfreturn variables.maxCachedRoutes>
	</cffunction>
	
	<cffunction name="setServerName" access="public" returntype="void" output="false">
		<cfargument name="value" type="string" required="true">

		<cfif not len(arguments.value) or value eq "cgi.server_name">
			<cfset arguments.value = cgi.server_name>
		</cfif>

		<cfset variables.serverName = arguments.value>
	</cffunction>
	
	<cffunction name="getServerName" access="public" returntype="any" output="false">
		
		<cfreturn variables.serverName>
	</cffunction>
	
	<cffunction name="setProtocol" access="public" returntype="void" output="false">
		<cfargument name="value" type="string" required="true">

		<cfset variables.protocol = arguments.value>
	</cffunction>
	
	<cffunction name="getProtocol" access="public" returntype="any" output="false">

		<cfif not len(variables.protocol) or variables.protocol eq "auto">
			<cfif cgi.https eq "on">
				<cfreturn "https">
			<cfelse>
				<cfreturn "http">
			</cfif>
		</cfif>

		<cfreturn variables.protocol>
	</cffunction>
	
	<cffunction name="setAbsoluteUrl" access="public" returntype="void" output="false">
		<cfargument name="value" type="boolean" required="true">

		<cfset variables.absoluteUrl = arguments.value>
	</cffunction>
	
	<cffunction name="getAbsoluteUrl" access="public" returntype="any" output="false">

		<cfreturn variables.absoluteUrl>
	</cffunction>
	
	<cffunction name="setScriptName" access="public" returntype="void" output="false">
		<cfargument name="value" type="string" required="true">

		<cfset variables.scriptName = arguments.value>
	</cffunction>
	
	<cffunction name="getScriptName" access="public" returntype="any" output="false">

		<cfif variables.scriptName eq "cgi.script_name">
			<cfreturn cgi.script_name>
		</cfif>
		
		<cfreturn variables.scriptName>
	</cffunction>
	
	<cffunction name="setAppendSlash" access="public" returntype="void" output="false">
		<cfargument name="value" type="boolean" required="true">

		<cfset variables.appendSlash = arguments.value>
	</cffunction>
	
	<cffunction name="getAppendSlash" access="public" returntype="any" output="false">

		<cfreturn variables.appendSlash>
	</cffunction>
	
	<cffunction name="setPrefix" access="public" returntype="void" output="false">
		<cfargument name="value" type="string" required="true">

		<cfset variables.prefix = arguments.value>
				
		<!--- Prefixes should begin with a / and not end with one --->
		<cfif len(variables.prefix)>
			<cfif right(variables.prefix,1) EQ "/">
				<cfset variables.prefix = mid( variables.prefix, 1, len(variables.prefix)-1 )>
			</cfif>
			<cfif left(variables.prefix,1) NEQ "/">
				<cfset variables.prefix = "/" & variables.prefix>
			</cfif>
		</cfif>
		
	</cffunction>
	
	<cffunction name="getPrefix" access="public" returntype="any" output="false">

		<cfreturn variables.prefix>
	</cffunction>

	<!---+
		Connects a route path and requirements for this mapper. A route is 
		an expression which can be matched against paths to generate parameters
		for a request.
		
		For instance the route "blog/:year/:month/:day" can be used to match
		requests for blog/2006/05/01. Requirements may be used to ensure that
		the matched parts of route always have a specific format. They may
		also be used to mark a group as optional, set a default or change
		case sensitivity.
		
		See the Route.cfc for information about how to construct a path
		expression and what params should be provided (paramRequirements).
				
		A route name may also be provided, If a route name is provided it can be
		used when generating routes later to avoid searching. Names must be 
		unique, if there's a duplicate route with the same name a
		accelerate.routing.DuplicateRouteName exception will be thrown.
		
		The newly connected route object is returned so it can be inspected.
	--->
	<cffunction name="connect" access="public" output="false" returntype="any">
		<cfargument name="path" type="string" required="true">
		<cfargument name="params" type="struct" required="false" default="#structNew()#">
		<cfargument name="name" type="string" required="false" default="">
		<cfargument name="static" type="boolean" required="false" default="false">
		
		<cfset var route = "">
		
		<cfif not arguments.static>
			<cfset route = createObject("component","Route").init( arguments.path, arguments.params )>
		<cfelse>
			<cfset route = arguments.path>
		</cfif>

		<!--- 
			If a route name was provided add the route to the named routes
			collection.
		--->
		<cfif len(arguments.name)>
			<cfif NOT structKeyExists(variables.namedRoutes,arguments.name)>
				<cfset variables.namedRoutes[arguments.name] = route>
			<cfelse>
				<cfthrow type="accelerate.routing.DuplicateRouteName" message="Duplicate route named '#arguments.name#'" detail="All route names must be unique">
			</cfif>
		</cfif>
		
		<cfif not arguments.static>
			<!--- 
				Add the route, don't create the search cache yet as there's no way
				know what terms will be used to search with later.
			--->
			<cfset arrayAppend( variables.routes, route )>
		</cfif>
		
		<cfreturn route>
	</cffunction>
	
	<!---
		Matches a path against the routes in this matcher. A strcuture with the 
		routing parameters will be returned. Two special parameters are returned
		as well, __route__ which is a reference to the route that matched, and
		__path__ has the actual URL that was matched.
		
		If no route can be matched false will be returned instead.
	--->
	<cffunction name="match" access="public" output="false" returntype="any">
		<cfargument name="path" type="string" required="true">
	
		<cfset var match = "">
		<cfset var cacheKey = "route:" & arguments.path>
		<cfset var result = "">
		<cfset var i = "">
				
		<!--- Check the cache for a cached route, if there is one use that... --->
		<cfset result = getCachedObject( cacheKey )>
		<cfif isStruct(result)>
			<cfreturn deepCopy(result)>
		</cfif>
	
		<!--- Strip out the prefix --->
		<cfset arguments.path = REReplace(path,"^" & variables.prefix,"")>
				
		<!--- ...Else look through all the routes for one that matches --->
		<cfloop from="1" to="#arrayLen(variables.routes)#" index="i">
			<cfset match = variables.routes[i].match( path )>
		
			<!--- 
				Break when route is matched. Routes are imutable so its
				safe to append the route object as __route__.
			--->
			<cfif isStruct(match)>
				<cfset match.__path__ = arguments.path>
				<cfset match.__route__ = variables.routes[i]>
				<cfset structAppend( match, variables.routingDefaults, false )>
				<cfset addCachedObject( cacheKey, match )>
				<cfreturn match>
			</cfif>
		</cfloop>

		<cfreturn false>
	</cffunction>
	
	<!---
		Generates a new path based on parameters that as passed in. The parameters
		must match a route in this matcher by having the same "extra" values which
		are not defined in the path expression, as well as having all required
		params for the expression.
		
		When generating the closest maching routes will always be checked first
		based on the param names. For instance if the search params:
		
			struct( controller="blog", year="2006" )
			
		are provided then then all routes that have only controller and year
		params will be checked first, this includes params that are not in the
		path expression and are "extra". Next all routes that have 1 differing
		parameter will be checked, and so on.
		
		False will be returned if no path can be generated.
		
		appendSlash:
			If a slash should be appended after the path. This should generally
			be left to false as it can have undesired effects on caching.
		
		scriptName:
			The name of the script to append if rewriting is not available. This
			is useful when using cgi.path_info; can be set to index.cfm and then
			all paths will be prefixed with that.
		
		absoluteUrl:
			If the url that is generated should be absolute. An absolute url has
			the protocol, serverName and scriptName appended to it.
			
		protocol:
			The protocol to use in the generated path. Only applied when the 
			absoluteUrl parameter is true. Defaults to http.
			
		serverName:
			The name of the server. Only applied when absoluteUrl is true.
			Defaults to the value of cgi.server_name
		
		name:
			The name of the route to use when generating the path.
	--->
	<cffunction name="generate" access="public" output="false" returntype="any">
		<cfargument name="searchParams" type="struct" required="false" default="#structNew()#">
		<cfargument name="appendSlash" type="boolean" required="false" default="#variables.appendSlash#">
		<cfargument name="scriptName" type="string" required="false" default="#getScriptName()#">
		<cfargument name="absoluteUrl" type="boolean" required="false" default="#variables.absoluteUrl#">
		<cfargument name="protocol" type="string" required="false" default="#getProtocol()#">
		<cfargument name="serverName" type="string" required="false" default="#getServerName()#">
		<cfargument name="name" type="string" required="false" default="">
		
		<cfset var searchParamNames = "">
		<cfset var searchRoutes = "">
		<cfset var i = "">
		<cfset var result = "">
		<cfset var params = structCopy( arguments.searchParams )>
		<cfset var cacheKey = "">
		<cfset var key = "">
		<cfset var generatedPath = "">
		<cfset var route = "">
				
		<!--- Remove special parameters in case this is from memory --->
		<cfset structDelete( params, "__route__" )>
		<cfset structDelete( params, "__path__" )>
				
		<!--- Make unique string for cache key from arguments --->
		<cfset cacheKey = arguments.toString()>
				
		<!--- Attempts to get the cached route for this path --->
		<cfset result = getCachedObject( cacheKey )>

		<!--- Return the cache if it exists --->
		<cfif not isBoolean(result)>
			<cfreturn result>
		</cfif>
		
		<!--- Use a named route if provided, otherwise search them all --->
		<cfif len(arguments.name)>
			<cfif structKeyExists( variables.namedRoutes, arguments.name )>
				<cfset route = variables.namedRoutes[arguments.name]>
				<cfif isObject(route)>
					<cfset result = route.generate( params, arguments.appendSlash )>
				<cfelse>
					<cfreturn route>
				</cfif>
			<cfelse>
				<cfthrow type="accelerate.routing.InvalidRouteName" message="No route by the named '#arguments.name#'">
			</cfif>
		<cfelse>
			<!--- Loop through all the routes searching for one that matches --->
			<cfloop from="1" to="#arrayLen(variables.routes)#" index="i">
				<cfset generatedPath = variables.routes[i].generate( params, arguments.appendSlash )>
				<cfif not isBoolean(generatedPath)>
					<cfset result = generatedPath>
					<cfbreak>
				</cfif>
			</cfloop>
		</cfif>
				
		<!---
			If a matching route was found append the extra parts and cache 
			the result. Then return the result.
		--->
		<cfif NOT isBoolean(result)>
			<cfif NOT len(arguments.name)>
				<cfset result = variables.prefix & result>

				<!--- Prevent double slashs --->
				<cfif right(arguments.scriptName,1) eq "/">
					<cfset arguments.scriptName = left(arguments.scriptName,len(arguments.scriptName)-1)>
				</cfif>
				
				<cfset result = arguments.scriptName & result>

				<cfif arguments.absoluteUrl>
					<cfset result = arguments.protocol & "://" & arguments.serverName & result>
				</cfif>

				<cfset addCachedObject( cacheKey, result )>
			</cfif>
			<cfreturn result>
		</cfif>

		<cfreturn false>
	</cffunction>
	
</cfcomponent>