<!---+
	Service that handles getting, creating and autowiring objects.
--->
<cfcomponent output="false" extends="AbstractService">
	
	<cffunction name="init" access="public" returntype="any" output="false">

		<cfset variables.objects = structNew()>
		<cfset variables.autowireCache = structNew()>

		<cfreturn this>
	</cffunction>
	
	<!---+
		Gets a shared object from the bean factory or creates an object and stores
		it as a singleton.
		
		If it creates an object that wasn't in the bean factory then it calls the 
		init method with no arguments. If you need constructor arguments use the
		bean factory instead.
		
		Example: getObject("models.TwitterService") turns into:
		
		// synchronized
		if( not structKeyExists(objects,arguments.name) ) {
			object = newObject(arguments.name).init()
			objects[arguments.name] = object;
		}
		return objects[arguments.name];
	--->
	<cffunction name="getObject" access="public" returntype="any" output="false">
		<cfargument name="name" type="any" required="true">
		
		<cfset var object = "">
				
		<cfif structKeyExists(variables.objects,arguments.name)>
			<cfreturn variables.objects[arguments.name]>
		
		<cfelseif getBeanFactory().containsBean(arguments.name)>
			<cfset object = getBeanFactory().getBean(arguments.name)>
			
			<cfif getBeanFactory().getBeanDefinition(arguments.name).isSingleton()>			
				<cfset variables.objects[arguments.name] = object>
			</cfif>
			
			<cfreturn object>
		</cfif>
		
		<!--- If it's not in CS then try to create it --->
		<cfif not structKeyExists(variables.objects,arguments.name)>
			<cflock timeout="30" name="Accelerate.NewObject(#getMapping()#.#arguments.name#)">
				<cfif not structKeyExists(variables.objects,arguments.name)>
					<!--- Always try making the model first --->
					<cfif fileExists(expandPath(
						replace("/#getMapping()#.models.#name#",".","/","all") & ".cfc"
					))>
						<cfset object = newObject("models." & arguments.name).init()>
					<cfelse>
						<cfset object = newObject(arguments.name).init()>
					</cfif>
					
					<cfset variables.objects[arguments.name] = object>														
				</cfif>
			</cflock>
		</cfif>
		
		<cfreturn variables.objects[arguments.name]>
	</cffunction>
	
	<!---+
		Creates a new object like createObject(component), but checks the app
		bundle first. So we could do newObject("models.Email") to get a new
		email object.
		
		Will not autowire the object if the cfcomponent tag has autowire="false".
	--->
	<cffunction name="newObject" access="public" returntype="any" output="false">
		<cfargument name="name" type="string" required="true">
		
		<cfset var qualifiedName = "#getMapping()#.#arguments.name#">
		<cfset var object = "">
		<cfset var error = "">
		
		<cfset logAppend("info","Framework","newObject()","Creating new instance of #name#")>
		
		<!--- First try the app bundle --->
		<cfif fileExists(expandPath("/#replace(qualifiedName,".","/","all")#.cfc"))>
			<cfset object = createObject("component",qualifiedName)>
		<cfelse>
			<cfset object = createObject("component",arguments.name)>
		</cfif>
		
		<cfif not structKeyExists(getMetaData(object),"autowire") or getMetaData(object).autowire>
			<cfset autowireObject(object)>
		</cfif>		
		
		<cfreturn object>
	</cffunction>	
	
	<!---+
		Autowires an object with the objects defined in the bean factory.
	--->
	<cffunction name="autowireObject" access="public" returntype="void" output="false">
		<cfargument name="object" required="true" type="any">

		<cfset var className = getMetaData(arguments.object).name>		
		<cfset var cacheEntry = "">
		<cfset var property = "">
		<cfset var bean = "">
		
		<cfif not structKeyExists(variables.autowireCache,className)>
			<cfset populateCache(object)>
		</cfif>
		
		<cfset cacheEntry = variables.autowireCache[className]>
		
		<cfloop collection="#cacheEntry#" item="property">
			<cfset bean = getObject(cacheEntry[property])>
			<cfinvoke component="#arguments.object#" method="#property#">
				<cfinvokeargument name="1" value="#bean#">
			</cfinvoke>
		</cfloop>
	</cffunction>
	
	<!---+
		Populates the autowire cache for an object instance.
	--->
	<cffunction name="populateCache" access="private" returntype="any" output="false">
		<cfargument name="object" type="any" required="true">
		
		<cfset var beanFactory = getBeanFactory()>
		<cfset var propertyName = "">
		<cfset var param = "">
		<cfset var method = "">
		<cfset var bean = "">
		<cfset var objectType = "">
		<cfset var objectName = "">
		<cfset var autowireCacheEntry = structNew()>
		<cfset var className = getMetaData(arguments.object).name>
		
		<cfloop collection="#arguments.object#" item="propertyName">
			<cfif isCustomFunction(arguments.object[propertyName])>
				<cfset method = getMetaData(arguments.object[propertyName])>
				<cfif len(method.name) gt 3 and left(method.name,3) eq "set" 
					  and arrayLen(method.parameters) eq 1>
					
					<cfset objectName = mid(method.name,4,len(method.name))>
					<cfset param = method.parameters[1]>
					
					<cfif beanFactory.containsBean(objectName)>
						<cfset bean = getObject(objectName)>
						
					<cfelseif structKeyExists(param,"type") 
						and not listFindNoCase("any,string,numeric,array,struct",param.type)>
						
						<cfset objectType = param.type>
						<cfset objectName = beanFactory.findBeanNameByType(objectType)>
						
						<cfif len(objectName)>
							<cfset bean = getObject(objectName)>
						</cfif>
					<cfelse>
						<cfset objectName = "">
					</cfif>
					
					<cfif len(objectName)>
						<cfset autowireCacheEntry[propertyName] = objectName>
					</cfif>
				</cfif>
			</cfif>
		</cfloop>
		
		<cfset variables.autowireCache[className] = autowireCacheEntry>
	</cffunction>
	
	<cffunction name="getMapping" access="private" returntype="string" output="false">
		
		<cfreturn getConfig().getMapping()>
	</cffunction>
	
</cfcomponent>